--[[ Alternia
     No, you are not being trolled.

]]--

local addoninfo, alt = ...
local lbag = Library.LibBaggotry
local acct = Library.LibAccounts
local filt = Library.LibEnfiltrate
local printf = Library.printf.printf
local printfhs = Library.printf.printfhs
local sprintf = Library.printf.sprintf
local html_safe = Library.printf.html_safe
local slower = string.lower
local supper = string.upper
local tconcat = table.concat
local tsort = table.sort
alt.version = "0.11-151128-15:56:44"

local function capitalize(name)
  name = name or "unknown"
  return supper(name:sub(1, 1)) .. slower(name:sub(2, -1))
end

function alt.variables_loaded(event, name)
  local shard_data = Inspect.Shard()
  local shard = shard_data and shard_data.name
  if name == 'Alternia' then
    AlterniaG = AlterniaG or {}
    AlterniaA = AlterniaA or {}
  end
  if shard then
    AlterniaA[shard] = AlterniaA[shard] or {}
    AlterniaG[shard] = AlterniaG[shard] or {}
  end
  alt.shard_data = AlterniaG[shard]
  if alt.whoami then
    AlterniaG[shard][alt.whoami] = AlterniaG[shard][alt.whoami] or { currency = {} }
    alt.data = AlterniaG[shard][alt.whoami]
  end
end

function alt.pretty_coin(n)
  local components = {}
  local s = n % 100
  n = (n - s) / 100
  local g = n % 100
  local p = (n - g) / 100
  if n == 0 then
    return '<font color="#888888">0p</font>'
  end
  if p > 0 then
    components[#components + 1] = sprintf('<font color="#ddddff">%dp</font>', p)
  end
  if g > 0 or p > 0 then
    if g > 0 then
      components[#components + 1] = sprintf('<font color="#bbaa22">%dg</font>', g)
    else
      components[#components + 1] = sprintf('<font color="#998811">0g</font>', g)
    end
  end
  if s > 0 or g > 0 or p > 0 then
    if s > 0 then
      components[#components + 1] = sprintf('<font color="#dddddd">%ds</font>', s)
    else
      components[#components + 1] = sprintf('<font color="#999999">0s</font>')
    end
  end
  return tconcat(components, "")
end

function alt.xpbar(xp)
  local xpbar = ""
  local progress = xp.accumulated / xp.needed
  local color = "cccccc"
  local accumulated, needed, rested
  if progress > .9 then
    color = "66ff66"
  elseif progress > .75 then
    color = "99dd66"
  elseif progress > .5 then
    color = "bbbb66"
  elseif progress > .25 then
    color = "dd9966"
  elseif progress > .1 then
    color = "ff6666"
  else
    color = "dd4444"
  end
  if xp.accumulated >= 10000000  then
    accumulated = sprintf("%.1fM", xp.accumulated / 1000000)
  elseif xp.accumulated >= 10000 then
    accumulated = sprintf("%.1fK", xp.accumulated / 1000)
  else
    accumulated = sprintf("%d", xp.accumulated)
  end
  if xp.rested and xp.rested > 0 then
    if xp.rested >= 1000000  then
      rested = sprintf("+%.1fM", xp.rested / 1000000)
    elseif xp.rested >= 1000 then
      rested = sprintf("+%dK", xp.rested / 1000)
    else
      rested = sprintf("+%d", xp.rested)
    end
  else
    rested = ""
  end
  if xp.needed >= 10000000  then
    needed = sprintf("/%.1fM", xp.needed / 1000000)
  elseif xp.needed >= 10000 then
    needed = sprintf("/%.1fK", xp.needed / 1000)
  else
    needed = sprintf("/%d", xp.needed)
  end
  xpbar = sprintf("<font color=\"#%s\">%s</font><font color=\"#9999cc\">%s</font>%s", color, accumulated, rested, needed)
  if xp.needed and xp.accumulated then
    if xp.rested and xp.rested > 0 then
      rested = sprintf("<font color=\"#9999cc\">+%d%%</font>", xp.rested / xp.needed * 100)
    else
      rested = ""
    end
    xpbar = xpbar .. sprintf(" (<font color=\"#%s\">%d%%</font>%s)", color, xp.accumulated / xp.needed * 100, rested)
  end
  return xpbar
end

function alt.slashcommand(args)
  local chars_to_process = {}
  if not args then
    return
  end
  if args.s then
    if lbag.share_inventory(not lbag.share_inventory_p()) then
      printf("Now sharing inventory with other accounts.")
    else
      printf("Inventory now private again.")
    end
    return
  end
  if args.v then
    printf("version %s", alt.version)
    return
  end
  -- we need to check which accounts characters are on, because we
  -- want to skip characters who aren't on this account for some currencies.
  if (args.c or args.l) and #args.leftover_args > 0 and alt.shard_data then
    local chars = {}
    local names = {}
    if not alt.shard_data then
      printf("No data for this shard, so no characters can match any names.")
      return
    end
    for i = 1, #args.leftover_args do
      names[#names + 1] = slower(args.leftover_args[i])
    end
    local found_any = false
    for char, data in pairs(alt.shard_data) do
      for i = 1, #names do
        if char:find(names[i]) then
	  found_any = true
	  chars[char] = true
	end
      end
    end
    if not found_any then
      printf("No characters matching provided names.")
      return
    end
    -- this loop is so we can have multiple name specifications which
    -- match a single character result in only one instance of that name,
    -- because it was being used as a hash key.
    for k, v in pairs(chars) do
      chars_to_process[#chars_to_process + 1] = k
    end
  elseif (args.c or args.l) then
    if alt.shard_data then
      for k, v in pairs(alt.shard_data) do
        chars_to_process[#chars_to_process + 1] = k
      end
    else
      printf("No data for this shard, so nothing to report.")
      return
    end
  end
  tsort(chars_to_process)
  if args.c then
    args.c = slower(args.c)
    local found = {}
    local found_any = false
    local colors_for = {}
    local this_acct = acct.acct_of(alt.whoami)
    for i = 1, #chars_to_process do
      local char = chars_to_process[i]
      local acct = acct.acct_of(char)
      local data = alt.shard_data[char]
      if data.currency then
        for k, v in pairs(data.currency) do
	  local skip = false
	  if v.bind == 'account' and acct ~= this_acct then
	    skip = true
	  end
	  if not skip and (k:find(args.c) or slower(v.name):find(args.c)) then
	    local name = v.name or k
	    local number = v.stack or 1
	    if not colors_for[name] then
	      local a, b, c = lbag.rarity_p(v.rarity or 'common', true)
	      if c then
	        colors_for[name] = sprintf("#%02x%02x%02x", c.r * 255, c.g * 255, c.b * 255)
	      else
	        colors_for[name] = "#dddddd"
	      end
	    end
	    local pretty_stack
	    if k == 'coin' then
	      pretty_stack = html_safe(alt.pretty_coin(number))
	    else
	      pretty_stack = sprintf("%d", number)
	      if #pretty_stack > 3 then
	        pretty_stack = pretty_stack:sub(1, -4) .. ',' .. pretty_stack:sub(-3, -1)
	      end
	    end
	    if number > 0 then
	      if not found_any then
		printf("Matching currency:")
	        found_any = true
	      end
	      found[name] = (found[name] or 0) + number
	      printfhs("  <font color=\"#ffffaa\">%s</font><font color=\"#aaaaaa\">:</font> %s <font color=\"%s\">%s</font>", capitalize(char), pretty_stack, colors_for[name], name)
	    end
	  end
	end
      end
    end
    if found_any then
      printf("Totals:")
      for k, v in pairs(found) do
	local pretty_stack
	if k == 'Platinum, Gold, Silver' then
	  pretty_stack = html_safe(alt.pretty_coin(v))
	else
	  pretty_stack = sprintf("%d", v)
	  if #pretty_stack > 3 then
	    pretty_stack = pretty_stack:sub(1, -4) .. ',' .. pretty_stack:sub(-3, -1)
	  end
	end
        printfhs("  %s <font color=\"%s\">%s</font>", pretty_stack, colors_for[k], k)
      end
    else
      printf("No currency matching '%s'.", args.c)
    end
  end
  if args.l then
    if #chars_to_process >= 1 then
      printf("Level advancement:")
    else
      printf("Found no characters.")
    end
    for i = 1, #chars_to_process do
      local char = chars_to_process[i]
      local data = alt.shard_data[char]
      did_rested = false
      if data.experience then
        local cap = capitalize(char)
	local lines = {}
	local level = 0
	if data.experience.xp then
	  level = data.experience.xp.level or 0
	  if data.experience.xp.needed then
	    lines[#lines + 1] = sprintf("Level <font color=\"#ffffff\">%d</font>", level)
	    lines[#lines + 1] = sprintf("%s", alt.xpbar(data.experience.xp))
	    if data.experience.xp.rested then
	      did_rested = true
	    end
	  else
	    lines[#lines + 1] = sprintf("Level <font color=\"#66ff66\">%d</font>", level)
	  end
	end
	-- only show PA levels for characters at/above 50.
        if data.experience.attunement and level >= 50 then
	  local a = data.experience.attunement
	  local color
	  if a.available == 0 then
	    color = "66ff66"
	  else
            color = "bbbb66"
	  end
	  lines[#lines + 1] = sprintf("<font color=\"#%s\">%d/%d</font> PA", color, a.spent, (a.spent + a.available))
	  if a.needed then
	    if did_rested then
	      -- don't print rest bonus twice
	      a = { needed = a.needed, accumulated = a.accumulated }
	    end
	    lines[#lines + 1] = sprintf("%s", alt.xpbar(a))
	  end
	end
	printfhs("  <font color=\"#eeeeee\">%s</font><font color=\"#bbbbbb\">: %s</font>", cap, html_safe(tconcat(lines, ", ")))
      end
    end
  end
end

function alt.me_first(a, b, mine)
  if mine[a] and not mine[b] then
    return true
  end
  if mine[b] and not mine[a] then
    return false
  end
  return a > b
end

function alt.tooltip_update(data)
  -- do nothing
  local maxwidth = 0
  local counter = 0
  local charcounts = {}
  local characters = {}
  local mine = {}
  local found_any = false
  for _, details in pairs(data) do
    local c = details._character or 'unknown'
    local s = details._slotspec or 'unknown'
    if c == alt.whoami then
      local ok, val = pcall(function() return Utility.Item.Slot.Parse(s) end)
      if ok then
        c = val or s or 'nil'
      else
        c = 'owned'
	-- We use the bogus slot spec "Wallet" for currency.
	-- printf("error: %s", type)
      end
      mine[c] = true
    end
    if charcounts[c] then
      charcounts[c] = charcounts[c] + (details.stack or 1)
    else
      charcounts[c] = details.stack or 1
      characters[#characters + 1] = c
    end
    found_any = true
  end
  if not found_any then
    return false
  end
  tsort(characters, function(a, b) return alt.me_first(a, b, mine) end)
  for _, c in ipairs(characters) do
    local cap = capitalize(c)
    local pretty_stack = sprintf("%d", charcounts[c])
    if #pretty_stack > 3 then
      pretty_stack = pretty_stack:sub(1, -4) .. ',' .. pretty_stack:sub(-3, -1)
    end
    local pretty = string.format("%s: %s", cap, pretty_stack)
    counter = counter + 1
    if counter > #alt.labels then
      table.insert(alt.labels, UI.CreateFrame('Text', sprintf("alternia_label_%d", counter), alt.subframe))
      alt.labels[counter]:SetPoint("BOTTOMRIGHT", alt.labels[counter - 1], "TOPRIGHT", 0, 1)
      alt.labels[counter]:SetFontSize(alt.labels[counter]:GetFontSize() + 2)
    end
    alt.labels[counter]:SetText(pretty)
    if mine[c] then
      alt.labels[counter]:SetFontColor(0.95, 0.95, 0.6)
    else
      alt.labels[counter]:SetFontColor(0.85, 0.85, 0.85)
    end
    alt.labels[counter]:SetVisible(true)
    local width = alt.labels[counter]:GetWidth()
    if width > maxwidth then
      maxwidth = width
    end
  end
  alt.subframe:SetPoint("TOP", alt.labels[counter], "TOP", nil, -2)
  alt.subframe:SetWidth(maxwidth + 4)
  if counter < #alt.labels then
    for i = counter + 1, #alt.labels do
      alt.labels[i]:SetVisible(false)
    end
  end
  return true
end

function alt.show_tooltip(data)
  if not alt.tooltip then
    alt.labels = {}
    alt.tooltip = UI.CreateFrame('Frame', 'alternia_tooltip_frame', alt.ui)
    if not alt.tooltip then
      printf("Couldn't create a tooltip.  Oops.")
      return
    end
    alt.tooltip:SetBackgroundColor(0.4, 0.4, 0.3, 1.0)
    alt.subframe = UI.CreateFrame('Frame', 'alternia_tooltip_subframe', alt.tooltip)
    alt.subframe:SetBackgroundColor(0.1, 0.1, 0.1, 1.0)
    table.insert(alt.labels, UI.CreateFrame('Text', 'alternia_tooltip_1', alt.subframe))

    alt.tooltip:SetPoint("BOTTOMLEFT", UI.Native.Tooltip, "TOPLEFT", 5, 5)
    alt.tooltip:SetPoint("TOPRIGHT", alt.subframe, "TOPRIGHT", 1, -1)
    alt.subframe:SetPoint("BOTTOMLEFT", alt.tooltip, "BOTTOMLEFT", 1, -1)
    alt.labels[1]:SetPoint("BOTTOMRIGHT", alt.subframe, "BOTTOMRIGHT", -2, -2)
    alt.labels[1]:SetFontSize(alt.labels[1]:GetFontSize() + 2)

    alt.labels[1]:SetText("")
  end
  alt.tooltip:SetVisible(alt.tooltip_update(data))
  local point, offset
  -- This attempts to ensure a little margin near edge of screen; note that
  -- the visible native tooltip is not the actual height of the window
  if UI.Native.Tooltip:GetTop() - alt.tooltip:GetHeight() < 0 then
    point = "BOTTOMLEFT"
    offset = alt.tooltip:GetHeight() + 5
  else
    point = "TOPLEFT"
    offset = 5
  end
  -- printf("Native's top %d, my height %d, point %s, offset %d.", UI.Native.Tooltip:GetTop(), alt.tooltip:GetHeight(), point, offset)
  alt.tooltip:SetPoint("BOTTOMLEFT", UI.Native.Tooltip, point, 5, offset)
end

function alt.hide_tooltip()
  if alt.tooltip then
    alt.tooltip:SetVisible(false)
  end
  Command.Event.Detach(Event.System.Update.Begin, nil, 'alternia_tooltip')
end

function alt.handle_tooltip(item)
  if Inspect.Time.Frame() == alt.this_frame then
    return
  end
  Command.Event.Detach(Event.System.Update.Begin, nil, 'alternia_tooltip')
  local found_any = false
  local found_items = {}
  if alt.shard_data then
    for char, data in pairs(alt.shard_data) do
      if data.currency and data.currency[item] then
        found_items[#found_items + 1] = data.currency[item]
      end
    end
  end
  if #found_items == 0 then
    local filter = filt.Filter:new(nil, 'item', 'Alternia')
    filter:require({ field = 'type', relation = '==', value = item })
    -- we will accept anything which is not bound, or anything which binds
    -- to account.
    filter:include({ field = 'bound', relation = '!=', value = true })
    filter:include({ field = 'bind', relation = '==', value = 'account' })
    lbag.apply_args(filter, { C = '*', s = 'all' })
    -- use expanded data
    found_items = lbag.expand(filter)
  end
  alt.show_tooltip(found_items)
end

function alt.tooltip_handler(event, tt_type, shown, buff)
  if tt_type == 'item' or tt_type == 'itemtype' then
    alt.this_frame = Inspect.Time.Frame()
    local ok, details = pcall(function() return Inspect.Item.Detail(shown) end)
    if details and details.type then
      -- remove any existing handler
      Command.Event.Detach(Event.System.Update.Begin, nil, 'alternia_tooltip')
      Command.Event.Attach(Event.System.Update.Begin, function() alt.handle_tooltip(details.type) end, 'alternia_tooltip')
    else
      if not ok then
        printf("Error: %s", details)
      end
      alt.hide_tooltip()
    end
  else
    if Inspect.Time.Frame() ~= alt.this_frame then
      alt.hide_tooltip()
    end
  end
end

function alt.startup_handler()
  alt.whoami = slower(Inspect.Unit.Detail('player').name)
  if alt.whoami and alt.shard_data then
    alt.shard_data[alt.whoami] = alt.shard_data[alt.whoami] or { currency = {} }
    alt.data = alt.shard_data[alt.whoami]
    alt.currency_handler()
  end
end

function alt.currency_handler()
  if alt.data then
    alt.data.currency = Inspect.Currency.Detail(Inspect.Currency.List())
    -- don't try to track credits, as they're meaningless
    alt.data.currency.credit = nil
    for k, v in pairs(alt.data.currency) do
      if not v.rarity then
        local status, result = pcall(Inspect.Item.Detail, v.id)
	if status then
	  v.rarity = result.rarity or 'common'
	else
	  v.rarity = common
	end
      end
      -- this is tragic
      if v.id ~= 'coin' then
        v.bind = 'account'
      end
      v._character = alt.whoami
      v._slotspec = 'Wallet'
    end
  end
end

function alt.player_handler(event, units)
  for k, v in pairs(units) do
    if v == 'player' and alt.shard_data then
      Command.Event.Detach(Event.Unit.Availability.Full, alt.player_handler)
      local xp = Inspect.Experience()
      local attunement = Inspect.Attunement.Progress()
      local details = Inspect.Unit.Detail(k)
      xp.level = details.level
      alt.shard_data[alt.whoami].experience = { xp = xp, attunement = attunement }
      Command.Event.Attach(Event.Attunement.Progress.Accumulated, alt.attunement_handler, "attunement tracking")
      Command.Event.Attach(Event.Attunement.Progress.Available, alt.attunement_handler, "attunement tracking")
      Command.Event.Attach(Event.Attunement.Progress.Rested, alt.attunement_handler, "attunement tracking")
      Command.Event.Attach(Event.Attunement.Progress.Spent, alt.attunement_handler, "attunement tracking")
      Command.Event.Attach(Event.Experience.Accumulated, alt.xp_handler, "xp tracking")
      Command.Event.Attach(Event.Experience.Rested, alt.xp_handler, "xp tracking")
      Command.Event.Attach(Event.Unit.Detail.Level, alt.xp_handler, "xp tracking")
    end
  end
end

function alt.attunement_handler(event, units)
  alt.shard_data[alt.whoami].experience.attunement = Inspect.Attunement.Progress()
end

function alt.xp_handler(event, units)
  local xp = Inspect.Experience()
  if xp then
    local details = Inspect.Unit.Detail('player')
    if details and details.level then
      xp.level = details.level
    end
    alt.shard_data[alt.whoami].experience.xp = xp
  end
end

Library.LibGetOpt.makeslash('c:lsv', "Alternia", "alt", alt.slashcommand)

alt.ui = UI.CreateContext("Alternia")
alt.ui:SetStrata('tutorial')

Command.Event.Attach(Event.Addon.SavedVariables.Load.End, alt.variables_loaded, "variable loaded hook")
Command.Event.Attach(Event.Addon.Startup.End, alt.startup_handler, "startup hook")
Command.Event.Attach(Event.Unit.Availability.Full, alt.player_handler, "player availability hook")
Command.Event.Attach(Event.Tooltip, alt.tooltip_handler, "tooltip hook")
Command.Event.Attach(Event.Currency, alt.currency_handler, "currency update hook")
